# -*- coding: utf-8 -*-

"""Main module."""


import glob
import pandas as pd
from .util import Progbar

try:
    import bs4
except ImportError as identifier:
    raise Exception('Please instell BeautifulSoup4 to for parse_html to work: pip install beautifulsoup4')

def parse_html(file, select={}, limit=None, autounbox=True, addMetaNames=True, metaNamesPrefix='meta_', tags=['h1','h2','h3','b','em']):
    """
    parse_html('./papers.nips.cc/paper/*.html', select={
    'title': 'title',
    'message': 'p.abstract',
    'author': 'li.author a'
    }).apply(year='_.meta_citation_publication_date')
    """
    files = glob.glob(file)
    rows = []
    progbar = Progbar(len(files))
    for i,fname in enumerate(files):
        cols = {}
        # print(f"Parsing file {fname}")
        if limit is not None and limit <= i:
            break
        with open(fname) as fp:
            soup = bs4.BeautifulSoup(fp)
        for k,v in select.items():
            x = soup.select(v)
            cols[k] = [_.get_text() for _ in x]
        rows.append(cols)
        if addMetaNames:
            for meta in soup.find_all('meta'):
                if 'name' in meta.attrs and 'content' in meta.attrs:
                    n = metaNamesPrefix+meta.attrs['name']
                    if n not in cols:
                        cols[n] = []
                    cols[n].append(meta.attrs['content'])
        cols['body'] = soup.get_text()
        for tag in tags:
            cols[tag] = [ _.get_text() for _ in soup.find_all(tag) ]
        cols['a'] = [ _['href'] for _ in soup.find_all('a') ]
        progbar.add(1)

    all_cols = set(k for _ in rows for k in _.keys())
    if autounbox:
        for col in all_cols:
            for r in rows:
                if col in r and len(r[col]) > 1:
                    break
            else:
                for r in rows:
                    if col in r and len(r[col]) == 1:
                            r[col] = r[col][0]
        lens = { k: len for k in all_cols }
        
    return pd.DataFrame(rows)



# -*- coding: utf-8 -*-

"""Main module."""


import pandas as pd
import spacy

from . import elastic
from . import kibana
from .util import chunker, Tictoc

class Pipeline(object):
    def __init__(self, index, textCols=None, tagCols=None, numCols=None, geoPointCols = None,
            idCol=None, dateCol=None, suggests=None, elk=None, lang='english', doctype='_doc'):
        self._pipeline = []
        self._index = index
        self._doctype=doctype
        self._textCols = textCols or []
        self._tagCols = tagCols or []
        self._numCols = numCols or []
        self._geoPointCols = geoPointCols or []
        self._ignoreUploadCols = []
        self._dateCol = dateCol
        self._idCol = idCol
        self._suggests = suggests or []
        self._lang = lang
        self.elk = elk
        self._tictoc = Tictoc(output='', additive=True)
        self._min_max = {}
    def add(self, x):
        self._pipeline.append(x)
        x.addingToPipeline(self)
    def __iadd__(self, other):
        self.add(other)
        return self
    def suggests(self, suggestCols):
        self._suggests = suggestCols

    def setup_elastic(self, **kwargs):
        if self.elk is not None:
            self.elk.createIndex(self._index, self._doctype,
                textCols=self._textCols, tagCols=self._tagCols,
                dateCol=self._dateCol, geoPointCols=self._geoPointCols, lang=self._lang, **kwargs)
    def setup_kibana(self, **kwargs):
        visCols = []
        if self._dateCol:
            visCols.append(kibana.DateHistogram(self._dateCol))
        for i in self._tagCols:
            visCols.append(kibana.HorizontalBar(i))
        for i in self._textCols:
            visCols.append(kibana.TagCloud(i))
        for i in self._numCols:
            interval = 0.1
            if i in self._min_max:
                min, max = self._min_max[i]
                interval = (max - min) / 100
            visCols.append(kibana.Histogram(i, interval))
        timeFrom, timeTo = None, None
        if self._dateCol in self._min_max:
            timeFrom, timeTo = self._min_max[self._dateCol]
        self.elk.kibana.setup_kibana(self._index, self._dateCol,
            searchCols=self._textCols, visCols=visCols, dashboard=True,
            timeFrom=timeFrom, timeTo=timeTo, **kwargs)
    def create_kibana_dashboard(self, **kwargs):
        self.setup_kibana(**kwargs)

    def process(self, texts, writeElastic=None, batchSize=1000, returnProcessed=True, progbar=True):
        if writeElastic is None:
            writeElastic = self.elk is not None
        if writeElastic:
            self.setup_elastic()
        results = []

        self.tic('global', 'process')
        for chunk in chunker(texts, batchSize, progbar=progbar):
            x = chunk
            for i,p in enumerate(self._pipeline):
                if not progbar:
                    print(f'Stage {i+1} of {len(self._pipeline)}: {p.name}')
                self.tic(f'Stage {i+1}',p.name)
                x = p.process(x)
                self.toc()
            if writeElastic:
                self.write_elastic(x, chunksize=batchSize, showProgbar=not progbar, setKibanaTimeDefault=False)
            if returnProcessed:
                results.append(x)
        self.toc()
        # return results
        self.tic('global', 'concat results')
        results =  pd.concat(results, sort=False)
        self.toc()

        self.tic('global','min_max_calc')
        # Need to keep track of ranges for kibana to have something ot work on
        cols = self._numCols
        if self._dateCol is not None:
            cols.append(self._dateCol)
        for i in cols:
            self._min_max[i] = (results.loc[:,i].min(),results.loc[:,i].max())
        self.toc()

        return results

    def write_elastic(self, texts, setKibanaTimeDefault=True, chunksize=1000, showProgbar=True):
        self.tic('elastic', 'upload')
        self.elk.loadDocs(index=self._index, doctype=self._doctype, dateCol=self._dateCol, idCol=self._idCol,
                            suggestCol=self._suggests, texts=texts.drop(columns=self._ignoreUploadCols, errors='ignore'), chunksize=chunksize, showProgbar=showProgbar)
        self.toc()
        if setKibanaTimeDefault and self._dateCol is not None:
            self.elk._kibana.set_kibana_timeDefaults(timeFrom=str(texts[self._dateCol].min()), timeTo=str(texts[self._dateCol].max()))
    def tic(self, part, name):
        self._tictoc.tic(f"{part} / {name}")
    def toc(self):
        self._tictoc.toc()

class PipelineStage(object):
    def __init__(self, name=None):
        self.name = type(self).__name__ if name is None else name
    def addingToPipeline(self, pipeline):
        self._pipeline = pipeline
    def doprocess(self, x):
        raise NotImplementedError()
    def tic(self, name):
        self._pipeline.tic(self.name, name)
    def ticwrap(self, iter, name):
        return self._pipeline._tictoc.wrap(iter, f"{self.name} / {name}")
    def toc(self):
        self._pipeline.toc()


class MapToSingle(PipelineStage):
    def __init__(self, col, outCol):
        super(MapToSingle, self).__init__()
        self._col = col if isinstance(col, str) else col[0]
        self._outCol = outCol
    def process(self, text):
        #print(self._col, type(text[self._col]))
        target = text[self._col].apply(lambda x: self.doprocess(str(x)))
        text = text.copy()
        text.loc[:,self._outCol] = target
        return text

class MapToTags(PipelineStage):
    def __init__(self, cols, outCol):
        super(MapToTags, self).__init__()
        self._cols = [cols] if isinstance(cols, str) else cols
        self._outCol = outCol
    def addingToPipeline(self, pipeline):
        super(MapToTags, self).addingToPipeline(pipeline)
        pipeline._tagCols.append(self._outCol)
    def process(self, text):
        target = []
        for i in range(len(text)):
            vals = []
            for c in self._cols:
                vals += (self.doprocess(text[c].iloc[i]))
            target.append(vals)
        text = text.copy()
        text.loc[:,self._outCol] = pd.Series(target)
        return text

class RegexTag(MapToTags):
    def __init__(self, regex='doi:[^ ]+', *args, **kwargs):
        super(RegexTag, self).__init__(*args, **kwargs)
        import re
        self._regex = re.compile(regex)
    def doprocess(self, x):
        y = self._regex.findall(x)
        return y

class VaderSentiment(MapToSingle):
    def __init__(self, *args, **kwargs):
        super(VaderSentiment, self).__init__(*args, **kwargs)
        from vaderSentiment.vaderSentiment import SentimentIntensityAnalyzer
        self._analyzer = SentimentIntensityAnalyzer()
    def addingToPipeline(self, pipeline):
        super(VaderSentiment, self).addingToPipeline(pipeline)
        pipeline._numCols.append(self._outCol)
    def doprocess(self, x):
        y = self._analyzer.polarity_scores(x)['compound']
        return y

# pipeline = Pipeline(index='nips', textCols=['message','title'])
# pipeline.add(RegexTag(r'\$([^$]+)\$', ['message'], 'math'))
# pipeline.add(VaderSentiment('message', 'sentiment'))
# x = pipeline.process(nips, writeElastic=False)
# x[['math','sentiment']].head(10)

class SynonymTags(MapToTags):
    # def __init__(self, ['Neural', 'Bayesian'], topn=10), ['message'], 'hypekeyword'):
        pass
class Split(MapToTags):
    # def __init__(self, ', '), ['author']):
        pass


###########
## spaCy ##
###########
class MapToNamedTags(PipelineStage):
    def __init__(self, cols, tags, ignoreUploadCols, coerceValsToStr=True):
        super(MapToNamedTags, self).__init__()
        self._cols = [cols] if isinstance(cols, str) else cols
        self._tags = tags
        self._outCols = [ c+'_'+k for c in self._cols for k in tags ]
        self._ignoreUploadCols = [ c+'_'+k for c in self._cols for k in ignoreUploadCols ]
        self._coerceValsToStr = coerceValsToStr
    def addingToPipeline(self, pipeline):
        super(MapToNamedTags, self).addingToPipeline(pipeline)
        # TODO discern between tagCols and numCols...
        pipeline._tagCols.extend(self._outCols)
        pipeline._ignoreUploadCols.extend(self._ignoreUploadCols)
    def process(self, text):
        target = [ text ]
        for c in self._cols:
            x = text[c]
            if self._coerceValsToStr:
                # WORKAROUND
                # spacy.pipe has a bug where '' gives an assertion error (unflatten) if
                # it is on the end of the pipeleine... however ' ' seems to be ok.
                x = x.fillna(' ').astype(str)
            y = self.doprocess(x)
            y = pd.DataFrame(y, index=text.index).rename(mapper=lambda x: c+'_'+x, axis=1)
            target.append(y)
        text = pd.concat( target, axis=1, sort=False)
        return text


class SpacyEnrichment(MapToNamedTags):
    # =spacy.load('en')
    def __init__(self, nlp='en_core_web_sm', *args, tags=['ents','subj','verb'], posNum=True, vec=False, returnDoc=False,
                batch_size=1000, n_threads=-1, **kwargs):
        super(SpacyEnrichment, self).__init__(*args, tags=tags, ignoreUploadCols=['doc','vec','vec_normalized'], **kwargs)
        self._nlp = spacy.load(nlp) if isinstance(nlp, str) else nlp
        self._posNum = posNum
        self._batch_size = batch_size
        self._n_threads = n_threads
        self._returnDoc = returnDoc
        self._vec = vec
    def doprocess(self, x):
        docs = []
        self.tic("spacy make iter")
        spacy_iter = self._nlp.pipe(x.values, batch_size=self._batch_size,
                                n_threads=self._n_threads)
        self.toc()
        for doc in self.ticwrap(spacy_iter, 'spacy iter'):
            self.tic('process spacy docs')
            ret = { 'wc': len(doc) }
            if ret['wc'] == 0:
                docs.append(ret)
                continue
            if 'ents' in self._tags:
    #             ents = pd.DataFrame({ 'ent': e.text, 'lab': e.label_} for e in doc.ents )
                for e in doc.ents:
                    ret['entity_'+e.label_] = ret.get('entity_'+e.label_, []) + [e.text]
                    ret['ents'] = ret.get('ents', []) + [e.text]

            tok = pd.DataFrame({ 'text': w.text, 'lemma': w.lemma_, 'pos': w.pos_, 'dep': w.dep_} for w in doc )
            if 'subj' in self._tags:
                ret['subj'] = list(tok.lemma.loc[tok.dep.isin(['nsubj','sb'])])
            if 'verb' in self._tags:
                ret['verb'] = list(tok.lemma.loc[tok.pos=='VERB'])

            posNum = tok.pos.value_counts()
            posSel = tok.pos.unique() if self._posNum==True else self._posNum
            for pos in posSel:
                ret[ 'num_'+pos ] = posNum.loc[pos]
            if self._vec:
                if self._vec == True or self._vec == 'unnormalized':
                    ret['vec'] = doc.vector
                if self._vec == True or self._vec == 'normalized':
                    ret['vec_normalized'] = doc.vector / doc.vector_norm
            if self._returnDoc:
                ret['doc'] = doc
            docs.append(ret)
            self.toc()
        return docs

def nlp_disp(doc, jupyter=True):
    '''Displays Spacy dependency trees (best in jupyter)'''
    return spacy.displacy.render(doc, style='dep', jupyter=jupyter)




# -*- coding: utf-8 -*-

"""Top-level package for NLPeasy."""

__author__ = """Philipp Thomann"""
__email__ = 'philipp.thomann@d-one.ai'
__version__ = '0.6.1'

from .pipeline import *
from .docker import *
from .elastic import *
from .html import *
from . import util



# -*- coding: utf-8 -*-

"""Main module."""
# TODO maybe rename this module to docker_util so we are distinct from the docker-py we use?


import glob
from pathlib import Path
import pandas as pd

from . import elastic

try:
    import docker as dockerpy
except ImportError as identifier:
    raise Exception('Please instell the python docker package for start_elastic_on_docker to work: pip install docker')

def start_elastic_on_docker(prefix, errorIfExists=False,
                            elkVersion='latest', elasticPassword='nlp is fun',
                            mountVolumePrefix=Path('./elastic-data/'), mountType='bind', mkdir=True,
                            kibanaPluginDirectory=None,
                            elasticPort=None, kibanaPort=None,
                            client=dockerpy.from_env(), forcePull=False, rm=True,
                            setAsDefaultElasticStack=True):
    """
    stack = start_elastic_on_docker('mynlpstack', version='6.3.2')
    """
    assert kibanaPluginDirectory is None

    el_im, ki_im = [ f"docker.elastic.co/{i}/{i}-oss:{elkVersion}" for i in ['elasticsearch','kibana'] ]
    if forcePull:
        client.images.pull(el_im)
        client.images.pull(ki_im)
    network = f'{prefix}_network'
    if not len(client.networks.list(names=[network])):
        client.networks.create(network)


    el_name = prefix+"_elastic"
    el_ulimits = [dockerpy.types.Ulimit(name='memlock', soft=-1, hard=-1), dockerpy.types.Ulimit(name='nofile', soft=65536, hard=65536)]
    el_env = [f'ELASTIC_PASSWORD={elasticPassword}','discovery.type=single-node',
             'cluster.name=docker-cluster','bootstrap.memory_lock=true','"ES_JAVA_OPTS=-Xms512m -Xmx512m"']
    if mountVolumePrefix is not None:
        p = Path(mountVolumePrefix)
        if not p.exists():
            if mkdir:
                p.mkdir(parents=False, exist_ok=False)
            else:
                raise FileNotFoundError(p)
        el_p = (p / 'elastic-data').absolute()
        ki_p = (p / 'kibana-plugins').absolute()
        el_p.mkdir(exist_ok=True)
        ki_p.mkdir(exist_ok=True)
        print(el_p, ki_p)
        el_mnts = [ dockerpy.types.Mount('/usr/share/elasticsearch/data', str(el_p), type=mountType ) ]
        ki_mnts = [ dockerpy.types.Mount('/usr/share/kibana/plugins', str(ki_p), type=mountType ) ]
    else:
        el_mnts, ki_mnts = [],[]
    if errorIfExists or not len(client.containers.list(filters={'name': el_name})):
        client.containers.run(el_im, name=el_name, detach=True, remove=rm, network=network, ports={'9200': elasticPort},
                environment=el_env, mounts=el_mnts, ulimits=el_ulimits)
    elasticPort = client.api.inspect_container(el_name)['NetworkSettings']['Ports']['9200/tcp'][0]['HostPort']

    ki_name = prefix+"_kibana"
    ki_env = [f'ELASTIC_PASSWORD={elasticPassword}',
             'SERVER_NAME=kibana', f'ELASTICSEARCH_URL=http://{el_name}:9200', f'ELASTICSEARCH_HOSTS=http://{el_name}:9200']
    if errorIfExists or not len(client.containers.list(filters={'name': ki_name})):
        client.containers.run(ki_im, name=ki_name, detach=True, remove=rm, network=network, ports={'5601': kibanaPort},
                environment=ki_env, mounts=ki_mnts)
    kibanaPort = client.api.inspect_container(ki_name)['NetworkSettings']['Ports']['5601/tcp'][0]['HostPort']

    elk = None
    # TODO find docker containers IP
    elk = elastic.ElasticStack(elasticPort=elasticPort, kibanaPort=kibanaPort)
    if setAsDefaultElasticStack:
        elastic.setDefaultStack(elk)

    return elk

def container_running(name, client=dockerpy.from_env(), raiseRuntimeError=None):
    c = [ _ for _ in client.containers.list(filters={'name': name}) if _.name == name ]
    if len(c) == 1 and c[0].status == 'running':
        return True
    if raiseRuntimeError:
        msg = raiseRuntimeError if isinstance(raiseRuntimeError, str) else f'container {name} not running'
        raise RuntimeError(msg)

def elasticStackFromDocker(containerPrefix, setAsDefaultStack=True, raiseRuntimeError=False):
    el_name, ki_name = containerPrefix+"_elastic", containerPrefix+"_kibana"
    client = dockerpy.from_env()
    if not container_running(el_name, client=client, raiseRuntimeError=raiseRuntimeError):
        return None
    elasticPort = client.api.inspect_container(el_name)['NetworkSettings']['Ports']['9200/tcp'][0]['HostPort']
    kibanaPort = client.api.inspect_container(ki_name)['NetworkSettings']['Ports']['5601/tcp'][0]['HostPort']
    from . import elastic
    es = elastic.ElasticStack(elasticPort=elasticPort, kibanaPort=kibanaPort)
    if setAsDefaultStack:
        elastic.setDefaultStack(es)
    return es


def stop_elastic_on_docker(containerPrefix):
    client = dockerpy.from_env()
    for c in client.containers.list(filters={'name': containerPrefix+'_'}):
        c.stop()
    network = f'{containerPrefix}_network'
    for n in client.networks.list(names=[network]):
        n.remove()




# -*- coding: utf-8 -*-

"""Kibana module."""


import pandas
import requests
import json
from . import util

class Kibana(object):
    def __init__(self, host='localhost', port=5601, protocol='http',
                verify_certs=True, **kwargs):
        self._host = host
        self._port = port
        self._protocol = protocol
        self._verify_certs = verify_certs
        self._defaultIndexPatternUID = None
        self._defaultSearchUID = None
    def kibanaUrl(self, path=""):
        # TODO maybe URLEncode path?
        if path and path[0] != '/':
            # TODO Warn about missing initial '/'?
            path = '/'+path
        return f"{self._protocol}://{self._host}:{self._port}{path}"
    def alive(self, verbose=True):
        resp = requests.head(self.kibanaUrl())
        return resp.status_code == 200

    def show_kibana(self, how=None, *args, **kwargs):
        if how is None:
            how = 'jupyter' if util.IS_JUPYTER else 'webbrowser'
            # TODO can we figure out "non-interactive" to put how='print' then?
        how = how if isinstance(how, list) else [how]
        url = self.kibanaUrl(*args, **kwargs)
        if 'print' in how:
            print(f"Open: {url}")
        if 'webbrowser' in how:
            import webbrowser
            webbrowser.open(url)
        if 'jupyter' in how or 'ipython' in how:
            from IPython.core.display import HTML
            return HTML(self._repr_html_())

    def __repr__(self):
        return f"Kibana on {self.kibanaUrl()}"

    def _repr_html_(self):
        return f"Kibana on <a href='{self.kibanaUrl()}'>{self.kibanaUrl()}</a>"

    def getKibanaSavedObjects(self, type='index-pattern', search=None, fields=None):
        type = '&type=' + type if type else ''
        search = '&search=' + search if search else ''
        fields = '&fields=' + fields if fields else ''
        resp = requests.get(self.kibanaUrl(f'/api/saved_objects/_find?{type}{search}{fields}'))
        resp.raise_for_status()
        result = resp.json()['saved_objects']
        return result
    def postKibanaSavedObject(self, type, attributes, id=None):
        body = { "attributes": attributes }
        id = "/"+id if id else ""
        result = requests.post(self.kibanaUrl(f'/api/saved_objects/{type}{id}?overwrite=true'), headers={"kbn-xsrf": "true"}, json=body)
        result.raise_for_status()
        # return result.json()
        return result.json()['id'], result.json()
    def deleteKibanaSavedObject(self, type, uid):
        u = self.kibanaUrl(f'/api/saved_objects/{type}/{uid}')
        resp = requests.delete(u, headers={"kbn-xsrf": "true"})
        resp.raise_for_status
        print(resp.json())
        return resp.json()

    def truncateKibanaSavedObjects(self, types=['dashboard','visualization','search','index-pattern'], search=None):
        for t in types:
            if search is not None and t=='index-pattern_________':
                continue
            objs = self.getKibanaSavedObjects(type=t, fields='name', search=search)
            print(f'deleting {len(objs)} objects of type {t}...')
            for i in objs:
                # print(i['id'])
                self.deleteKibanaSavedObject(t, i['id'])
        print('finished deleting')

    def getKibanaConfig(self, name=None, onlyLastSetValue=True, defaultValue=None):
        assert onlyLastSetValue
        config = self.getKibanaSavedObjects('config')
        # TODO need to implement for onlyLastSetValue=False as well or warn if multiple values?
        result = dict()
        for i in config:
            c = i['attributes']
            result.update(c)
        if name is not None:
            if name in result:
                return result[name]
            else:
                return defaultValue
        return result

    def addKibanaConfig(self, name, value, addToList=False, id=None):
        assert not addToList
        attributes = { name: value }
        res = self.postKibanaSavedObject('config', attributes, id=id)
        return res
    def addKibanaIndexPattern(self, indexPattern, timeField=None, setDefaultIndexPattern=True, overwrite=False):
        if not overwrite:
            # TODO handle in postKibana URL parameter?
            self.getKibanaSavedObjects('index-pattern', indexPattern)
            for i in self.getKibanaSavedObjects('index-pattern',indexPattern,'title'):
                if i['attributes']['title'] == indexPattern:
                    return
        attributes = {
            "title": indexPattern,
        }
        if timeField is not None:
            attributes["timeFieldName"] = timeField
        uid, result = self.postKibanaSavedObject('index-pattern', attributes)
        if setDefaultIndexPattern:
            self._defaultIndexPatternUID = uid

        return uid, result
    def addKibanaSearch(self, title, columns, description=None, sort=None, setDefaultSearch=True, indexPatternUID=None):
        if indexPatternUID is None:
            indexPatternUID = self._defaultIndexPatternUID
        searchSourceJSON = {
            "index": indexPatternUID,
            # "highlightAll": True,
            # "version": True,
            "query":{"query":"","language":"kuery"},
            "filter":[]
        }
        attributes = {
            "title": title, 'columns': columns,
            'kibanaSavedObjectMeta': {'searchSourceJSON': json.dumps(searchSourceJSON)}
        }
        if description is not None:
            attributes['description'] = description
        if sort is not None:
            attributes['sort'] = sort
        uid, res = self.postKibanaSavedObject(type='search', attributes=attributes)
        if setDefaultSearch:
            self._defaultSearchUID = uid
        return uid, res
    def addVisualization(self, title, viz, indexPatternUID=None):
        if indexPatternUID is None:
            indexPatternUID = self._defaultIndexPatternUID
        # visState = {
        #     'aggs':[
        #         {'id': '1', 'schema':'metric', 'type': 'count'},
        #         {
        #             'id': '2', 'schema':'segment', 'type': 'terms',
        #             'params': {'field': field, 'size': size, 'order': 'desc', 'orderBy': '1',  }
        #         },
        #     ],
        #     'params': {
        #         'type': 'histogram'
        #     },
        #     'title': title,
        #     'type': visType,
        # }
        assert isinstance(viz, Visualization)
        visState = viz.visState(title)
        searchSourceJSON = {
            "index":indexPatternUID,
            "filter":[],
            "query":{"language":"kuery","query":""}
        }
        uid, res = self.postKibanaSavedObject('visualization', attributes={
            'title': title, 'visState': json.dumps(visState), 'uiStateJSON': '{"vis":{"legendOpen":false}}',
            'kibanaSavedObjectMeta': {'searchSourceJSON': json.dumps(searchSourceJSON)}
        })
        return uid, res
    def addDashboard(self, title, searchUID, visUIDs, timeFrom=None, timeTo=None, nVisCols=3, visW=16, visH=16, searchW=48, searchH=16):
        panels = [{
            'panelIndex': '1',
            'gridData': {'x': 0, 'y': 0, 'w': searchW, 'h': searchH, 'i': '1'},
            'version': '6.3.2',
            'type': 'search',
            'id': searchUID,
            'embeddableConfig': {}
        }]
        for i, v in enumerate(visUIDs):
            ix, iy = i % nVisCols, i // nVisCols
            x, y = ix*visW , searchH + iy*visH
            # print(ix,iy, x,y)
            iStr = str(i+2)
            panels.append({
                'panelIndex': iStr,
                'gridData': {'x': x, 'y': y, 'w': visW, 'h': visH, 'i': iStr},
                'version': '6.3.2',
                'type': 'visualization',
                'id': v,
                'embeddableConfig': {}
            })
        attributes = {
            'title': title,
        #      'hits': 0,
            'description': '',
            'panelsJSON': json.dumps(panels),
            'optionsJSON': '{"darkTheme":false,"useMargins":true,"hidePanelTitles":false}',
        #      'version': 1,
        #      'refreshInterval': {'display': 'Off', 'pause': False, 'value': 0},
            'kibanaSavedObjectMeta': {'searchSourceJSON': '{"query":{"query":"","language":"kuery"},"filter":[],"highlightAll":true,"version":true}'}
        }
        if timeFrom is not None and timeTo is not None:
            attributes['timeRestore'] = True
            attributes['timeTo'] = str(timeTo)
            attributes['timeFrom'] = str(timeFrom)
        uid, res = self.postKibanaSavedObject('dashboard',attributes)
        return uid, res

    def setup_kibana(self, index, timeField=None, searchCols=[], visCols=None, dashboard=True, timeFrom=None, timeTo=None, sets=True):
        print(f'{index}: adding index-pattern')
        ipUID, _ipRes = self.addKibanaIndexPattern(index, timeField, overwrite=True)
        if self.getKibanaConfig('defaultIndex') is None:
            # BUG the following is not really setting the defaultIndex as the Kibana UI see it...
            print(f'{index}: setting default index-pattern')
            self.addKibanaConfig('defaultIndex', ipUID)
        print(f'{index}: adding search')
        seUID, _seRes = self.addKibanaSearch(index+"-search", searchCols)
        visUIDs = []
        for i in visCols:
            if isinstance(i, str):
                i = HorizontalBar(i)
            print(f'{index}: adding visualisation for {i.field}')
            uid, _res = self.addVisualization(f'[{index}] {i.field}', i)
            visUIDs.append(uid)
        if dashboard:
            print(f'{index}: adding dashboard')
            daUID, _daRes = self.addDashboard(f'[{index}] Dashboard', seUID, visUIDs, timeFrom=timeFrom, timeTo=timeTo)
        if sets:
            print(f'{index}: setting time defaults')
            self.set_kibana_timeDefaults(timeFrom, timeTo)
        return {'index-patterh': ipUID, 'search': seUID, 'visualization': visUIDs, 'dashboard': daUID}

    def set_kibana_timeDefaults(self, timeFrom="now-15m", timeTo='now', mode='quick'):
        '''
        For accepted formats see https://www.elastic.co/guide/en/elasticsearch/reference/6.7/common-options.html#date-math
        '''
        # more configs on https://www.elastic.co/guide/en/kibana/current/advanced-options.html
        # maybe also set the timepicker:quickRanges key to a list of interesting time ranges.

        # value is a JSON, but as string.
        # btw we need to escape the outer {} in the f'...' string
        timeFrom = timeFrom or 'now-15m'
        timeTo = timeTo or 'now'
        value = { "from": str(timeFrom), "to": str(timeTo), "mode": "{mode}" }
        uid, res = self.addKibanaConfig("timepickerts", json.dumps(value))
        return uid, res
    def set_kibana_timeQuickRange(self, display, timeFrom, timeTo, section=3, id=None):
        '''
        For accepted formats see https://www.elastic.co/guide/en/elasticsearch/reference/6.7/common-options.html#date-math
        '''
        if id is None:
            result = requests.get(self.kibanaUrl('api/status'))
            result.raise_for_status()
            id = result.json()['version']['number']
        timeFrom = timeFrom or 'now-15m'
        timeTo = timeTo or 'now'
        value = [{"from": str(timeFrom), "to": str(timeTo), "display": display, "section": section}]
        uid, res = self.addKibanaConfig("timepicker:quickRanges", json.dumps(value), id=id)
        return uid, res
    def show_kibana_jupyter(self, height=500):
        # see e.g. https://github.com/tensorflow/tensorboard/blob/d9092143511cb04e4bfc904820305f1be45c67b3/tensorboard/notebook.py
        from IPython.display import IFrame
        url = self.kibanaUrl()
        iframe = IFrame(src=url, height=500, width="100%")
        return iframe

class Visualization(object):
    def __init__(self, field, agg='count'):
        self.field = field
        self.agg = agg
        self.visType = None
    def visState(self, title):
        visState = {
            'aggs':[
                {'id': '1', 'schema':'metric', 'type': self.agg},
                self.agg2(),
            ],
            'params': {
                'type': self.visType
            },
            'title': title,
            'type': self.visType,
        }
        return visState
    def agg2(self):
        raise NotImplementedError()
class HorizontalBar(Visualization):
    def __init__(self, field, size=20):
        super(HorizontalBar, self).__init__(field)
        self.visType = 'horizontal_bar'
        self.size = size
    def agg2(self):
        return {
                    'id': '2', 'schema':'segment', 'type': 'terms',
                    'params': {'field': self.field, 'size': self.size, 'order': 'desc', 'orderBy': '1',  }
                }
class TagCloud(HorizontalBar):
    def __init__(self, field, size=20):
        super(TagCloud, self).__init__(field, size)
        self.visType = 'tagcloud'
        self.params =  {
            "scale": "linear",
            "orientation": "single",
            "minFontSize": 18,
            "maxFontSize": 72,
            "showLabel": True
        }
class Histogram(Visualization):
    def __init__(self, field, interval):
        super(Histogram, self).__init__(field)
        self.visType = 'histogram'
        self.interval = interval
    def agg2(self):
        return {
            "id": "2",
            "enabled": True,
            "type": "histogram",
            "schema": "segment",
            "params": {
                "field": self.field,
                "interval": self.interval,
                "extended_bounds": {}
            }
        }
class DateHistogram(Visualization):
    def __init__(self, field):
        super(DateHistogram, self).__init__(field)
        self.visType = 'histogram'
    def agg2(self):
        return {
            "id": "2",
            "enabled": True,
            "type": "date_histogram",
            "schema": "segment",
            "params": {
                "field": self.field,
                "interval": "auto",
                "customInterval": "2h",
                "min_doc_count": 1,
                "extended_bounds": {}
            }
        }



from enum import Enum

class Lang(Enum):
    ## Format is: 2-letter-code = ( ElasticCode, [spacymodels] )
    DE = ('german', ['de_core_news_sm','de_core_news_md'])
    EL = ('greek', ['el_core_news_sm','el_core_news_md'])
    EN = ('english', ['en_core_web_sm','en_core_web_md','en_core_web_lg','en_vectors_web_lg'])
    ES = ('spanish', ['es_core_news_sm','es_core_news_md'])
    FR = ('french', ['fr_core_news_sm','fr_core_news_md'])
    IT = ('italian', ['it_core_news_sm'])
    NL = ('dutch', ['nl_core_news_sm','nl_core_news_md'])
    PT = ('portuguese', ['pt_core_news_sm'])
    XX = (None, ['xx_ent_wiki_sm'])

    def __init__(self, elastic, spacy_models):
        self.code = self.name
        self.elastic = elastic
        self.spacy_models = spacy_models
    @property
    def spacy_models_installed(self):
        from importlib.util import find_spec
        return [ i for i in self.spacy_models if find_spec(i) is not None ]



# -*- coding: utf-8 -*-

"""Main module."""


import pandas as pd
import elasticsearch
from . import kibana
from . import docker

from .util import Progbar, chunker, print_or_display, rmNanFromDict

def connect_elastic(dockerPrefix='nlp', startOnDocker=True,
            host='localhost', elasticPort=None, kibanaPort=None, kibanaHost=None,
            elkVersion='7.1.1', mountVolumePrefix=None,
            verbose=True, failOnNotAvailable=False, **kwargs):
    kibanaHost = kibanaHost or host
    elasticPort = elasticPort or 9200
    kibanaPort = kibanaPort or 5601
    log = print_or_display if verbose else lambda x: None
    elk = ElasticStack(host=host, elasticPort=elasticPort, kibanaPort=kibanaPort, kibanaHost=kibanaHost)
    if elk.alive():
        log(f"Elasticsearch already running")
        # TODO warn if version mismatches elkVersion param
        log(elk)
        return elk
    if dockerPrefix is None:
        if failOnNotAvailable:
            raise RuntimeError(f"No running elasticsearch found on {host}:{elasticPort}.")
        else:
            log(f"No running elasticsearch found on {host}:{elasticPort}.")
            return None

    # Let's pass it on to docker:
    log(f"No elasticsearch on {host}:{elasticPort} found, trying connect to docker container with prefix {dockerPrefix}")
    elk = docker.elasticStackFromDocker(containerPrefix=dockerPrefix, setAsDefaultStack=False)
    if elk is None or not elk.alive():
        if startOnDocker:
            log(f"No docker container with prefix {dockerPrefix}; starting one")
            assert all(i not in kwargs for i in ['mountVolumePrefix','version','prefix'])
            elk = docker.start_elastic_on_docker(prefix=dockerPrefix, elkVersion=elkVersion, mountVolumePrefix=mountVolumePrefix, **kwargs)
        else:
            msg = f"No running elasticsearch found on docker with prefix {dockerPrefix}."
            if failOnNotAvailable:
                raise RuntimeError(msg)
            else:
                log(msg)
                return None
    log(elk)
    return elk

class ElasticStack(object):
    def __init__(self, host='localhost', elasticPort=9200, kibanaPort=5601, protocol='http',
                kibanaHost=None, kibanaProtocol=None, verify_certs=True, setAsDefaultStack=True, **kwargs):
        self._host = host
        self._elasticPort = elasticPort
        self._protocol = protocol
        self._verify_certs = verify_certs

        self._es = None
        self._kibana = None
        self._elasticKwargs = kwargs

        self.kibana = kibana.Kibana(
            host=self._host if kibanaHost is None else kibanaHost,
            port=kibanaPort,
            protocol=self._protocol if kibanaProtocol is None else kibanaProtocol,
            verify_certs=self._verify_certs
        )

        if setAsDefaultStack:
            setDefaultStack(self)

    def alive(self, verbose=True):
        import logging
        urllib_logger = logging.getLogger("request")
        orig_max_retries = 3
        orig_level = urllib_logger.level
        result = False
        try:
            # BUG Disabling logging does not work:
            urllib_logger.setLevel(logging.FATAL)
            orig_max_retries = self.es.transport.max_retries
            self.es.transport.max_retries = 0
            result = self.es.ping()
        except Exception as e:
            if verbose:
                print(e)
        self.es.transport.max_retries = orig_max_retries
        urllib_logger.setLevel(orig_level)
        return result

    def url(self):
        return f"{self._protocol}://{self._host}:{self._elasticPort}"
    def __repr__(self):
        return f"ElasticSearch on {self.url()}\n" + self.kibana.__repr__()
    def _repr_html_(self):
        return f"ElasticSearch on <a href='{self.url()}'>{self.url()}</a> <br> " + self.kibana._repr_html_()


    @property
    def es(self):
        if self._es is None:
            host = { 'host': self._host, 'port': self._elasticPort, 'use_ssl': self._protocol == 'https',  }
            self._es = elasticsearch.Elasticsearch([host], verify_certs=self._verify_certs, **self._elasticKwargs)
        return self._es

    def getAnalysis(self, lang='english', synonyms=None):
        filter_names = []
        if lang == 'english':
            filter_names.append("english_possessive_stemmer")
        filter_names.append('lowercase')
        filters = {
            f"{lang}_stop": {
                "type": "stop",
                "stopwords": f"_{lang}_"
            },
            f"{lang}_stemmer": {
                "type": "stemmer",
                "language": f"{lang}"
            },
        }
        filter_names.extend(filters.keys())
        if lang == 'english':
            filters["english_possessive_stemmer"] = {
                "type": "stemmer",
                "language": "possessive_english"
            }
        if synonyms is not None:
            filters[f"{lang}_synonym"] = {
                "type": "synonym",
                "synonyms": synonyms
                # "synonyms_path": "analysis/synonym.txt"
            }
            filter_names.append(f"{lang}_synonym")

        analyzer = {
                    f"{lang}_syn": {
                                "tokenizer": "standard",
                                "filter": filter_names
                            }

        }
        return filters, analyzer


    # TODO languages, synonyms,
    def createIndex(self, index='texts',doctype='_doc',create=True,
            textCols=[], tagCols=[], geoPointCols = [], synonyms=[], dateCol=None, lang='english',
            deleteOld=True, verbose=False):
        # assert lang == 'english'
        properties = {}
        for k in textCols:
            # TODO Make sure that the analyzer is created as f"{lang}_syn":
            properties[k] =  { "type": "text", "fielddata": True, "analyzer": f"{lang}_syn" }
        for k in tagCols:
            properties[k] =  { "type": "keyword" }
        for k in geoPointCols:
            properties[k] =  { "type": "geo_point" }
        properties["suggest"] = { "type" : "completion" }
        mapping = {
            #"_timestamp": {"enabled": "false"},
            "properties": properties
        }
        if self.es.info()['version']['number'] < '7':
            mapping = { doctype: mapping }
        if create:
            filters, analyzer = self.getAnalysis(lang, synonyms)
            body={
                "settings": {
                    "analysis": {
                        "filter": filters,
                        # {
                        #     "synonym": {
                        #         "type": "synonym",
                        #         "synonyms": synonyms
                        #         # "synonyms_path": "analysis/synonym.txt"
                        #     }
                        # },
                        "analyzer": analyzer,
                        # "analyzer": {
                        #     "english_syn": {
                        #         "tokenizer": "standard",
                        #         "filter": [
                        #             "english_possessive_stemmer",
                        #             "lowercase",
                        #             "english_stop",
                        #             "english_stemmer",
                        #             "synonym"
                        #         ]
                        #     }
                        # }
                    }
                },

                "mappings": mapping
            }
            if verbose:
                print(body)
            if deleteOld:
                try:
                    self.es.indices.delete(index)
                except:
                    pass
            self.es.indices.create(index=index, body=body) # , ignore=[]
            return(body)
        else:
            self.es.indices.put_mapping(index=index, doc_type=doctype, body=mapping)

    def loadDocs(self, index, texts, doctype='_doc', dateCol=None, deleteOld=False, chunksize=1000, idCol=None,
                suggestCol=None, showProgbar=True):
        if idCol is None:
            idCol = texts.index
        if deleteOld:
            try:
                self.es.indices.delete(index)
            except:
                pass
        #createIndex(index=index, create=deleteOld)

        for ic, cdf in enumerate(chunker(texts, chunksize, progbar=showProgbar)):
            docs = cdf.to_dict(orient='records')
            for ii, doc in enumerate(docs):
                i = ic * chunksize + ii
                doc = rmNanFromDict(doc)
                if suggestCol and suggestCol in doc:
                    doc['suggest'] = doc[suggestCol]
                try:
                    self.es.index(index=index, doc_type=doctype, id=idCol[i], body=doc)
                except elasticsearch.ElasticsearchException as ex:
                    print(ex)
                    print(doc)
                    print('=' * 80)

    def truncate(self, index, doctype='text'):
        self._es.delete_by_query(index, {
            "query" : {
                "match_all" : {}
            }
        })

    def show_kibana(self, how=None, *args, **kwargs):
        self.kibana.show_kibana(how=how, *args, **kwargs)

__DEFAULT_STACK = None
def defaultStack():
    if __DEFAULT_STACK is None:
        __DEFAULT_STACK = ElasticStack()
    return __DEFAULT_STACK
def setDefaultStack(es):
    global __DEFAULT_STACK
    __DEFAULT_STACK = es




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

# with open('HISTORY.md') as history_file:
#     history = history_file.read()

requirements = [ 'pandas', 'spacy', 'elasticsearch', 'beautifulsoup4', 'vaderSentiment', 'docker', ]

setup_requirements = ['pytest-runner', 'twine', ]

test_requirements = ['pytest', 'scikit-learn', ]

setup(
    author="Philipp Thomann",
    author_email='philipp.thomann@d-one.ai',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        # "Programming Language :: Python :: 2",
        # 'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.4',
        # 'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Text Processing :: Markup :: HTML',
    ],
    description="Easy Peasy Language Squeezy",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + '\n\n', # + history,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='nlpeasy',
    name='nlpeasy',
    packages=find_packages(include=['nlpeasy']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/d-one/nlpeasy',
    version='0.6.1',
    zip_safe=False,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `NLPeasy` package."""

import pytest


import nlpeasy as ne
import pandas as pd

@pytest.fixture(scope="session")
def elk():
    """The ELK stack class to use.

    Reuse the same elk among the whole testing session by setting scope="session"
    """
    elk = ne.connect_elastic(dockerPrefix='nlp', elkVersion='7.4.0', mountVolumePrefix=None)
    return elk


def test_end_to_end(elk):
    """Sample pytest test function with the pytest fixture as an argument."""
    # read data as Pandas data frame
    # nips = pd.read_pickle("data_raw/nips.pickle")
    from sklearn.datasets import fetch_20newsgroups
    news_raw = fetch_20newsgroups()
    news = pd.DataFrame({'group': [news_raw['target_names'][i] for i in news_raw['target']],
                         'message': news_raw['data']
                         })

    # setup stages in the NLP pipeline and set textfields
    pipeline = ne.Pipeline(index='news', textCols=['message'], elk=elk)

    # pipeline += ne.RegexTag(r'\$([^$]+)\$', ['message'], 'math')
    # pipeline += ne.VaderSentiment('message', 'sentiment')
    # pipeline += ne.SpacyEnrichment(cols=['message','title'])

    N = 1000
    # do the pipeline
    news_enriched = pipeline.process(news.head(N), writeElastic=True)

    assert news_enriched.shape[0] == N

    # Create Kibana Dashboard of all the columns
    pipeline.create_kibana_dashboard()

    # open Kibana in webbrowser
    #elk.show_kibana()



# from keras.utils.generic_utils import Progbar

import sys
import collections
import time
import numpy as np

def clip_read():
    from pandas.io import clipboard
    clipboard.clipboard_get()
def clip_write(x):
    from pandas.io import clipboard
    clipboard.clipboard_set(x)

def chunker(seq, size, progbar=True):
    "Use this as: for batch in chunker(mylist, 1000): ..."
    # from http://stackoverflow.com/a/434328
    n = len(seq)
    if progbar:
        pbar: Progbar = Progbar(n)
        pbar.update(0)
    for pos in range(0, n, size):
        yield seq[pos:pos + size]
        if progbar:
            pbar.update(min(pos+size,n))
def insert_with_progbar(engine, df, name, if_exists='replace', chunksize=1000, **kwargs):
    # from https://stackoverflow.com/questions/39494056/progress-bar-for-pandas-dataframe-to-sql
    pbar = Progbar(len(df))
    for i, cdf in enumerate(chunker(df, chunksize)):
        replace = if_exists if i == 0 else "append"
        cdf.to_sql(name=name, con=engine, if_exists=replace, **kwargs)
        pbar.add(chunksize)

# Source?
class Progbar(object):
    """Displays a progress bar.

    # Arguments
        target: Total number of steps expected, None if unknown.
        width: Progress bar width on screen.
        verbose: Verbosity mode, 0 (silent), 1 (verbose), 2 (semi-verbose)
        stateful_metrics: Iterable of string names of metrics that
            should *not* be averaged over time. Metrics in this list
            will be displayed as-is. All others will be averaged
            by the progbar before display.
        interval: Minimum visual progress update interval (in seconds).
    """

    def __init__(self, target, width=30, verbose=1, interval=0.05,
                 stateful_metrics=None):
        self.target = target
        self.width = width
        self.verbose = verbose
        self.interval = interval
        if stateful_metrics:
            self.stateful_metrics = set(stateful_metrics)
        else:
            self.stateful_metrics = set()

        self._dynamic_display = ((hasattr(sys.stdout, 'isatty') and
                                  sys.stdout.isatty()) or
                                 'ipykernel' in sys.modules)
        self._total_width = 0
        self._seen_so_far = 0
        self._values = collections.OrderedDict()
        self._start = time.time()
        self._last_update = 0

    def update(self, current, values=None):
        """Updates the progress bar.

        # Arguments
            current: Index of current step.
            values: List of tuples:
                `(name, value_for_last_step)`.
                If `name` is in `stateful_metrics`,
                `value_for_last_step` will be displayed as-is.
                Else, an average of the metric over time will be displayed.
        """
        values = values or []
        for k, v in values:
            if k not in self.stateful_metrics:
                if k not in self._values:
                    self._values[k] = [v * (current - self._seen_so_far),
                                       current - self._seen_so_far]
                else:
                    self._values[k][0] += v * (current - self._seen_so_far)
                    self._values[k][1] += (current - self._seen_so_far)
            else:
                # Stateful metrics output a numeric value.  This representation
                # means "take an average from a single value" but keeps the
                # numeric formatting.
                self._values[k] = [v, 1]
        self._seen_so_far = current

        now = time.time()
        # info = ' - %.0fs' % (now - self._start)
        info = ' - %s' % (formatTime(now - self._start))
        if self.verbose == 1:
            if (now - self._last_update < self.interval and
                    self.target is not None and current < self.target):
                return

            prev_total_width = self._total_width
            if self._dynamic_display:
                sys.stdout.write('\b' * prev_total_width)
                sys.stdout.write('\r')
            else:
                sys.stdout.write('\n')

            if self.target is not None and self.target>0:
                try:
                    numdigits = int(np.floor(np.log10(self.target))) + 1
                except:
                    numdigits=1
                barstr = '%%%dd/%d [' % (numdigits, self.target)
                bar = barstr % current
                prog = float(current) / self.target
                prog_width = int(self.width * prog)
                if prog_width > 0:
                    bar += ('=' * (prog_width - 1))
                    if current < self.target:
                        bar += '>'
                    else:
                        bar += '='
                bar += ('.' * (self.width - prog_width))
                bar += ']'
            else:
                bar = '%7d/Unknown' % current

            self._total_width = len(bar)
            sys.stdout.write(bar)

            if current:
                time_per_unit = (now - self._start) / current
            else:
                time_per_unit = 0
            if self.target is not None and current < self.target:
                eta = time_per_unit * (self.target - current)
                eta_format = formatTime(eta)

                info = ' - ETA: %s' % eta_format
            else:
                if time_per_unit >= 1:
                    info += ' %.0fs/step' % time_per_unit
                elif time_per_unit >= 1e-3:
                    info += ' %.0fms/step' % (time_per_unit * 1e3)
                else:
                    info += ' %.0fus/step' % (time_per_unit * 1e6)

            for k in self._values:
                info += ' - %s:' % k
                if isinstance(self._values[k], list):
                    avg = np.mean(
                        self._values[k][0] / max(1, self._values[k][1]))
                    if abs(avg) > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                else:
                    info += ' %s' % self._values[k]

            self._total_width += len(info)
            if prev_total_width > self._total_width:
                info += (' ' * (prev_total_width - self._total_width))

            if self.target is not None and current >= self.target:
                info += '\n'

            sys.stdout.write(info)
            sys.stdout.flush()

        elif self.verbose == 2:
            if self.target is None or current >= self.target:
                for k in self._values:
                    info += ' - %s:' % k
                    avg = np.mean(
                        self._values[k][0] / max(1, self._values[k][1]))
                    if avg > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                info += '\n'

                sys.stdout.write(info)
                sys.stdout.flush()

        self._last_update = now

    def add(self, n, values=None):
        self.update(self._seen_so_far + n, values)

def formatTime(eta):
    if eta > 3600:
        eta_format = ('%d:%02d:%02d' %
                    (eta // 3600, (eta % 3600) // 60, eta % 60))
    elif eta > 60:
        eta_format = '%d:%02d' % (eta // 60, eta % 60)
    else:
        eta_format = '%ds' % eta
    return eta_format

def formatTime_ns(eta):
    if eta > 10e9:
        eta_format = formatTime(eta//1e9)
    elif eta > 1e9:
        eta_format = '%.1fs' % (eta/1e9)
    elif eta > 10e6:
        eta_format = '%dms' % (eta / 1e6)
    elif eta > 1e6:
        eta_format = '%.1fms' % (eta / 1e6)
    else:
        eta_format = '%.1fus' % (eta / 1e3)
    return eta_format


if 'time_ns' in dir(time):
    def _time_ns():
        return time.time_ns()
else:
    def _time_ns():
        return int(time.time() * 1e9)


class Tictoc(object):
    def __init__(self, output='print', additive=False):
        self.stack = []
        self.output = output
        from collections import defaultdict
        self._summarizer = defaultdict(int) if additive else False
        self._prefix = None
    def tic(self, name):
        if self._summarizer != False:
            self._summarizer[name]
        self.stack.append( (name, _time_ns()) )
    def toc(self):
        stop = _time_ns()
        name, start = self.stack.pop()
        dur = stop - start
        if self.output == 'print':
            timeString = formatTime_ns(dur)
            print(f"{name}: {timeString}")
        if self._summarizer != False:
            self._summarizer[name] += dur
    def clear(self):
        self.stack = []
    def wrap(self, iter, name):
        self.tic(name)
        for i in iter:
            self.toc()
            yield i
            self.tic(name)
        self.toc()
    def summary(self):
        if len(self.stack):
            print(f"Warning: stack is not empty but has {len(self.stack)} items")
        for k,v in self._summarizer.items():
            print(f"{k}: {formatTime_ns(v)}")

def rmNanFromDict(x):
    import pandas as pd
    if isinstance(x, dict):
        y = {}
        for k,v in x.items():
            if isinstance(v, list) or isinstance(v, dict):
                y[k] = rmNanFromDict(v)
            elif not pd.isna(v):
                y[k] = v
        return y
    if isinstance(x, list):
        y = []
        for v in x:
            if isinstance(v, list) or isinstance(v, dict):
                y.append(rmNanFromDict(v))
            elif not pd.isna(v):
                y.append(v)
        return y
    raise Exception("x has to be a list or a dict")

def _display_and_is_jupyter():
    try:
        from IPython import display, get_ipython
        if not get_ipython():
            raise ImportError()
        return (display.display, get_ipython().has_trait('kernel'))
    except ImportError:
        return (print, False)
print_or_display, IS_JUPYTER = _display_and_is_jupyter()
